from __future__ import with_statement

try:
    import simplejson as json
except ImportError:
    import json

import base64
import datetime
import unittest

from flask import g

from flask_peewee.rest import RestAPI, RestResource, Authentication, UserAuthentication
from flask_peewee.tests.base import FlaskPeeweeTestCase
from flask_peewee.tests.test_app import User, Message, Note, TestModel, APIKey
from flask_peewee.utils import get_next, make_password, check_password


class RestApiTestCase(FlaskPeeweeTestCase):
    def setUp(self):
        super(RestApiTestCase, self).setUp()
        TestModel.drop_table(True)
        APIKey.drop_table(True)
        APIKey.create_table()
        TestModel.create_table()
    
    def response_json(self, response):
        return json.loads(response.data)
    
    def auth_headers(self, username, password):
        return {'Authorization': 'Basic %s' % base64.b64encode('%s:%s' % (username, password))}
            
    def conv_date(self, dt):
        return dt.strftime('%Y-%m-%d %H:%M:%S')
    
    def assertAPIResponse(self, resp_json, body):
        self.assertEqual(body, resp_json['objects'])
    
    def assertAPIMeta(self, resp_json, meta):
        self.assertEqual(meta, resp_json['meta'])
    
    def assertAPIUser(self, json_data, user):
        self.assertEqual(json_data, {
            'username': user.username,
            'active': user.active,
            'join_date': self.conv_date(user.join_date),
            'admin': user.admin,
            'id': user.id,
        })
    
    def assertAPIUsers(self, json_data, users):
        for json_item, user in zip(json_data['objects'], users):
            self.assertAPIUser(json_item, user)
    
    def assertAPINote(self, json_data, note):
        self.assertEqual(json_data, {
            'user': note.user.id,
            'message': note.message,
            'created_date': self.conv_date(note.created_date),
            'id': note.id,
        })
    
    def assertAPINotes(self, json_data, notes):
        for json_item, note in zip(json_data['objects'], notes):
            self.assertAPINote(json_item, note)
    
    def assertAPIMessage(self, json_data, message):
        self.assertEqual(json_data, {
            'user': message.user.id,
            'content': message.content,
            'pub_date': self.conv_date(message.pub_date),
            'id': message.id,
        })
    
    def assertAPIMessages(self, json_data, messages):
        for json_item, message in zip(json_data['objects'], messages):
            self.assertAPIMessage(json_item, message)
        
    def assertAPITestModel(self, json_data, tm):
        self.assertEqual(json_data, {
            'data': tm.data,
            'id': tm.id,
        })
    
    def assertAPITestModels(self, json_data, tms):
        for json_item, tm in zip(json_data['objects'], tms):
            self.assertAPITestModel(json_item, tm)


class RestApiBasicTestCase(RestApiTestCase):
    def get_users_and_notes(self):
        users = self.create_users()
        
        notes = []
        for i in range(10):
            for user in users:
                notes.append(Note.create(user=user, message='%s-%s' % (user.username, i)))
        return users, notes
    
    def test_pagination(self):
        users, notes = self.get_users_and_notes()
        
        # do a simple list of the first 20 items
        resp = self.app.get('/api/note/?ordering=id')
        resp_json = self.response_json(resp)
        
        # verify we have page and link to next page
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '',
            'next': '/api/note/?ordering=id&page=2',
            'page': 1,
        })
        
        # verify response objects are paginated properly
        self.assertAPINotes(resp_json, notes[:20])
        
        # do a list of first 10 items
        resp = self.app.get('/api/note/?ordering=id&limit=10')
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '',
            'next': '/api/note/?ordering=id&limit=10&page=2',
            'page': 1,
        })
        
        # verify response objects are paginated properly
        self.assertAPINotes(resp_json, notes[:10])
        
        # grab the second page
        resp = self.app.get(resp_json['meta']['next'])
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '/api/note/?ordering=id&limit=10&page=1',
            'next': '/api/note/?ordering=id&limit=10&page=3',
            'page': 2,
        })
        
        # verify response objects are paginated properly
        self.assertAPINotes(resp_json, notes[10:20])
        
        # grab the last page
        resp = self.app.get(resp_json['meta']['next'])
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '/api/note/?ordering=id&limit=10&page=2',
            'next': '',
            'page': 3,
        })
        
        # verify response objects are paginated properly
        self.assertAPINotes(resp_json, notes[20:])
    
    def test_filtering(self):
        users, notes = self.get_users_and_notes()
        
        # do a simple filter on a related model
        resp = self.app.get('/api/note/?user=%s&ordering=id' % self.normal.id)
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '',
            'next': '',
            'page': 1,
        })
        self.assertAPINotes(resp_json, self.normal.note_set.order_by('id'))
        
        # do a filter following a join
        resp = self.app.get('/api/note/?user__username=admin&ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '',
            'next': '',
            'page': 1,
        })
        self.assertAPINotes(resp_json, self.admin.note_set.order_by('id'))
        
        # filter multiple fields
        notes = list(self.admin.note_set.order_by('id'))
        third_id = notes[3].id
        
        resp = self.app.get('/api/note/?user__username=admin&id__lt=%s&ordering=id' % third_id)
        resp_json = self.response_json(resp)
        self.assertAPINotes(resp_json, notes[:3])
        
        # do a filter using multiple values
        resp = self.app.get('/api/note/?user__username=admin&user__username=inactive&ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPIMeta(resp_json, {
            'model': 'note',
            'previous': '',
            'next': '',
            'page': 1,
        })
        self.assertAPINotes(resp_json, Note.filter(user__in=[self.admin, self.inactive]).order_by('id'))
    
    def test_filter_with_pagination(self):
        users, notes = self.get_users_and_notes()
        notes = list(self.admin.note_set.order_by('id'))
        
        # do a simple filter on a related model
        resp = self.app.get('/api/note/?user__username=admin&limit=4&ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPINotes(resp_json, notes[:4])
        
        next_url = resp_json['meta']['next']
        resp = self.app.get(next_url)
        resp_json = self.response_json(resp)
        
        self.assertAPINotes(resp_json, notes[4:8])
        
        next_url = resp_json['meta']['next']
        resp = self.app.get(next_url)
        resp_json = self.response_json(resp)
        
        self.assertEqual(resp_json['meta']['next'], '')
        self.assertAPINotes(resp_json, notes[8:])
        
        prev_url = resp_json['meta']['previous']
        resp = self.app.get(prev_url)
        resp_json = self.response_json(resp)
        
        self.assertAPINotes(resp_json, notes[4:8])
        
        prev_url = resp_json['meta']['previous']
        resp = self.app.get(prev_url)
        resp_json = self.response_json(resp)
        
        self.assertEqual(resp_json['meta']['previous'], '')
        self.assertAPINotes(resp_json, notes[:4])


class RestApiUserAuthTestCase(RestApiTestCase):
    def setUp(self):
        super(RestApiUserAuthTestCase, self).setUp()
        self.create_users()
    
    def create_notes(self):
        notes = [
            Note.create(user=self.admin, message='admin'),
            Note.create(user=self.normal, message='normal'),
        ]
        self.admin_note, self.normal_note = notes
        return notes
        
    def test_list_get(self):
        resp = self.app.get('/api/note/')
        resp_json = self.response_json(resp)
        
        self.assertAPIResponse(resp_json, [])
        self.assertAPIMeta(resp_json, {'model': 'note', 'next': '', 'page': 1, 'previous': ''})
        
        self.create_notes()
        
        resp = self.app.get('/api/note/?ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPINotes(resp_json, [
            self.admin_note,
            self.normal_note,
        ])
    
    def test_detail_get(self):
        resp = self.app.get('/api/note/1/')
        self.assertEqual(resp.status_code, 404)
        
        self.create_notes()
        
        resp = self.app.get('/api/note/%s/' % self.normal_note.id)
        resp_json = self.response_json(resp)
        self.assertAPINote(resp_json, self.normal_note)

    def test_auth_create(self):
        note_data = {'message': 'test', 'user': self.inactive.id}
        serialized = json.dumps(note_data)
        
        # this request is not authorized
        resp = self.app.post('/api/note/', data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.post('/api/note/', data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database
        resp = self.app.post('/api/note/', data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
    
    def test_create(self):
        note_data = {'message': 'test', 'user': self.inactive.id}
        serialized = json.dumps(note_data)
        
        # authorized as an admin
        resp = self.app.post('/api/note/', data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        new_note = Note.get(message='test')
        self.assertEqual(new_note.user, self.inactive)
        
        resp_json = self.response_json(resp)
        self.assertAPINote(resp_json, new_note)
    
    def test_auth_edit(self):
        self.create_notes()
        
        note_data = {'message': 'edited'}
        serialized = json.dumps(note_data)
        
        url = '/api/note/%s/' % self.admin_note.id
        
        # this request is not authorized
        resp = self.app.put(url, data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
    
    def test_edit(self):
        self.create_notes()
        
        note_data = {'message': 'edited'}
        serialized = json.dumps(note_data)
        
        url = '/api/note/%s/' % self.admin_note.id
        
        # authorized as an admin
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        note = Note.get(id=self.admin_note.id)
        self.assertEqual(note.message, 'edited')
        
        resp_json = self.response_json(resp)
        self.assertAPINote(resp_json, note)
    
    def test_auth_delete(self):
        self.create_notes()
        
        url = '/api/note/%s/' % self.admin_note.id
        
        # this request is not authorized
        resp = self.app.delete(url)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.delete(url, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database
        resp = self.app.delete(url, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
    
    def test_delete(self):
        self.create_notes()
        
        url = '/api/note/%s/' % self.admin_note.id
        
        # authorized as an admin
        resp = self.app.delete(url, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        self.assertEqual(Note.select().count(), 1)
        
        resp_json = self.response_json(resp)
        self.assertEqual(resp_json, {'deleted': 1})


class RestApiOwnerAuthTestCase(RestApiTestCase):
    def setUp(self):
        super(RestApiOwnerAuthTestCase, self).setUp()
        self.create_users()
    
    def create_messages(self):
        messages = [
            Message.create(user=self.admin, content='admin'),
            Message.create(user=self.normal, content='normal'),
        ]
        self.admin_message, self.normal_message = messages
        return messages
        
    def test_list_get(self):
        resp = self.app.get('/api/message/')
        resp_json = self.response_json(resp)
        
        self.assertAPIResponse(resp_json, [])
        self.assertAPIMeta(resp_json, {'model': 'message', 'next': '', 'page': 1, 'previous': ''})
        
        self.create_messages()
        
        resp = self.app.get('/api/message/?ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPIMessages(resp_json, [
            self.admin_message,
            self.normal_message,
        ])
    
    def test_detail_get(self):
        resp = self.app.get('/api/message/1/')
        self.assertEqual(resp.status_code, 404)
        
        self.create_messages()
        
        resp = self.app.get('/api/message/%s/' % self.normal_message.id)
        resp_json = self.response_json(resp)
        self.assertAPIMessage(resp_json, self.normal_message)

    def test_auth_create(self):
        message_data = {'content': 'test'}
        serialized = json.dumps(message_data)
        
        # this request is not authorized
        resp = self.app.post('/api/message/', data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.post('/api/message/', data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database
        resp = self.app.post('/api/message/', data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
    
    def test_create(self):
        message_data = {'content': 'test'}
        serialized = json.dumps(message_data)
        
        # authorized as an admin
        resp = self.app.post('/api/message/', data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        new_message = Message.get(content='test')
        self.assertEqual(new_message.user, self.normal)
        
        resp_json = self.response_json(resp)
        self.assertAPIMessage(resp_json, new_message)
    
    def test_auth_edit(self):
        self.create_messages()
        
        message_data = {'content': 'edited'}
        serialized = json.dumps(message_data)
        
        url = '/api/message/%s/' % self.normal_message.id
        
        # this request is not authorized
        resp = self.app.put(url, data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database, but not owner
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 403)
        
        # authorized, user in database, is owner
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        obj = Message.get(id=self.normal_message.id)
        self.assertEqual(obj.content, 'edited')
    
    def test_edit(self):
        self.create_messages()
        
        message_data = {'content': 'edited'}
        serialized = json.dumps(message_data)
        
        url = '/api/message/%s/' % self.normal_message.id
        
        # authorized as normal
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        message = Message.get(id=self.normal_message.id)
        self.assertEqual(message.content, 'edited')
        
        resp_json = self.response_json(resp)
        self.assertAPIMessage(resp_json, message)
    
    def test_auth_delete(self):
        self.create_messages()
        
        url = '/api/message/%s/' % self.normal_message.id
        
        # this request is not authorized
        resp = self.app.delete(url)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.delete(url, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database, not owner
        resp = self.app.delete(url, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 403)
        
        # authorized, user in database, is owner
        resp = self.app.delete(url, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
    
    def test_delete(self):
        self.create_messages()
        
        url = '/api/message/%s/' % self.normal_message.id
        
        # authorized as an admin
        resp = self.app.delete(url, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 200)
        
        self.assertEqual(Message.select().count(), 1)
        
        resp_json = self.response_json(resp)
        self.assertEqual(resp_json, {'deleted': 1})


class RestApiAdminAuthTestCase(RestApiTestCase):
    def test_list_get(self):
        resp = self.app.get('/api/user/')
        resp_json = self.response_json(resp)
        
        self.assertAPIResponse(resp_json, [])
        self.assertAPIMeta(resp_json, {'model': 'user', 'next': '', 'page': 1, 'previous': ''})
        
        self.create_users()
        
        resp = self.app.get('/api/user/?ordering=id')
        resp_json = self.response_json(resp)
        
        self.assertAPIUsers(resp_json, [
            self.admin,
            self.normal,
        ])
    
    def test_detail_get(self):
        resp = self.app.get('/api/user/1/')
        self.assertEqual(resp.status_code, 404)
        
        self.create_users()
        
        resp = self.app.get('/api/user/%s/' % self.normal.id)
        resp_json = self.response_json(resp)
        self.assertAPIUser(resp_json, self.normal)
        
        resp = self.app.get('/api/user/%s/' % self.inactive.id)
        self.assertEqual(resp.status_code, 404)

    def test_auth_create(self):
        self.create_users()
        
        new_pass = make_password('test')
        
        user_data = {'username': 'test', 'password': new_pass}
        serialized = json.dumps(user_data)
        
        # this request is not authorized
        resp = self.app.post('/api/user/', data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.post('/api/user/', data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database, but not an administrator
        resp = self.app.post('/api/user/', data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized as an admin
        resp = self.app.post('/api/user/', data=serialized, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
    
    def test_create(self):
        self.create_users()
        
        new_pass = make_password('test')
        
        user_data = {'username': 'test', 'password': new_pass}
        serialized = json.dumps(user_data)
        
        # authorized as an admin
        resp = self.app.post('/api/user/', data=serialized, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
        
        new_user = User.get(username='test')
        self.assertTrue(check_password('test', new_user.password))
        
        resp_json = self.response_json(resp)
        self.assertAPIUser(resp_json, new_user)
    
    def test_auth_edit(self):
        self.create_users()
        
        user_data = {'username': 'edited'}
        serialized = json.dumps(user_data)
        
        url = '/api/user/%s/' % self.normal.id
        
        # this request is not authorized
        resp = self.app.put(url, data=serialized)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database, but not an administrator
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized as an admin
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
    
    def test_edit(self):
        self.create_users()
        
        user_data = {'username': 'edited'}
        serialized = json.dumps(user_data)
        
        url = '/api/user/%s/' % self.normal.id
        
        # authorized as an admin
        resp = self.app.put(url, data=serialized, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
        
        user = User.get(id=self.normal.id)
        self.assertEqual(user.username, 'edited')
        
        resp_json = self.response_json(resp)
        self.assertAPIUser(resp_json, user)
    
    def test_auth_delete(self):
        self.create_users()
        
        url = '/api/user/%s/' % self.normal.id
        
        # this request is not authorized
        resp = self.app.delete(url)
        self.assertEqual(resp.status_code, 401)
        
        # authorized, but user does not exist in database
        resp = self.app.delete(url, headers=self.auth_headers('xxx', 'xxx'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized, user in database, but not an administrator
        resp = self.app.delete(url, headers=self.auth_headers('normal', 'normal'))
        self.assertEqual(resp.status_code, 401)
        
        # authorized as an admin
        resp = self.app.delete(url, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
    
    def test_delete(self):
        self.create_users()
        
        url = '/api/user/%s/' % self.normal.id
        
        # authorized as an admin
        resp = self.app.delete(url, headers=self.auth_headers('admin', 'admin'))
        self.assertEqual(resp.status_code, 200)
        
        self.assertEqual(User.select().count(), 2)
        
        resp_json = self.response_json(resp)
        self.assertEqual(resp_json, {'deleted': 1})


class RestApiKeyAuthTestCase(RestApiTestCase):
    def setUp(self):
        super(RestApiKeyAuthTestCase, self).setUp()
        
        self.tm1 = TestModel.create(data='test1')
        self.tm2 = TestModel.create(data='test2')
        
        self.k1 = APIKey.create(key='k', secret='s')
        self.k2 = APIKey.create(key='k2', secret='s2')
    
    def test_list_get(self):
        with self.flask_app.test_client() as c:
            resp = c.get('/api/testmodel/')
            self.assertEqual(resp.status_code, 401)
            self.assertEqual(g.api_key, None)
            
            resp = c.get('/api/testmodel/?key=k&secret=s2')
            self.assertEqual(resp.status_code, 401)
            self.assertEqual(g.api_key, None)
            
            resp = c.get('/api/testmodel/?key=k&secret=s')
            self.assertEqual(g.api_key, self.k1)
            resp_json = self.response_json(resp)

            self.assertAPITestModels(resp_json, [
                self.tm1,
                self.tm2,
            ])
            self.assertAPIMeta(resp_json, {'model': 'testmodel', 'next': '', 'page': 1, 'previous': ''})
    
    def test_create(self):
        with self.flask_app.test_client() as c:
            test_data = {'data': 't3'}
            serialized = json.dumps(test_data)
            
            resp = c.post('/api/testmodel/', data=serialized)
            self.assertEqual(resp.status_code, 401)
            self.assertEqual(g.api_key, None)
            
            resp = c.post('/api/testmodel/?key=k&secret=s2', data=serialized)
            self.assertEqual(resp.status_code, 401)
            self.assertEqual(g.api_key, None)

            # test passing in via get args
            resp = c.post('/api/testmodel/?key=k&secret=s', data=serialized)
            self.assertEqual(g.api_key, self.k1)
            resp_json = self.response_json(resp)
            
            self.assertEqual(TestModel.select().count(), 3)
            self.assertEqual(resp_json['data'], 't3')



import datetime

from flask import Flask, request, redirect, url_for, render_template, flash, g, Response

from peewee import *

# flask-peewee bindings
from flask_peewee.admin import Admin, ModelAdmin, AdminPanel
from flask_peewee.auth import Auth, BaseUser
from flask_peewee.db import Database
from flask_peewee.filters import QueryFilter
from flask_peewee.rest import RestAPI, RestResource, RestrictOwnerResource, UserAuthentication, AdminAuthentication, APIKeyAuthentication
from flask_peewee.utils import get_object_or_404, object_list, make_password


class TestFlask(Flask):
    def update_template_context(self, context):
        ret = super(TestFlask, self).update_template_context(context)
        self._template_context.update(context)
        return ret


app = TestFlask(__name__)
app.config.from_object('flask_peewee.tests.test_config.Configuration')

db = Database(app)

@app.before_request
def clear_context():
    app._template_context = {}


class User(db.Model, BaseUser):
    username = CharField()
    password = CharField()
    email = CharField()
    join_date = DateTimeField(default=datetime.datetime.now)
    active = BooleanField(default=True)
    admin = BooleanField(default=False, verbose_name='Can access admin')

    def __unicode__(self):
        return self.username
    
    def message_count(self):
        return self.message_set.count()


class Message(db.Model):
    user = ForeignKeyField(User)
    content = TextField()
    pub_date = DateTimeField(default=datetime.datetime.now)
    
    def __unicode__(self):
        return '%s: %s' % (self.user, self.content)


class Note(db.Model):
    user = ForeignKeyField(User)
    message = TextField()
    created_date = DateTimeField(default=datetime.datetime.now)


class TestModel(db.Model):
    data = TextField()
    
    class Meta:
        ordering = ('id',)


class AModel(db.Model):
    a_field = CharField()

class BModel(db.Model):
    a = ForeignKeyField(AModel)
    b_field = CharField()

class CModel(db.Model):
    b = ForeignKeyField(BModel)
    c_field = CharField()
    
class DModel(db.Model):
    c = ForeignKeyField(CModel)
    d_field = CharField()

class BDetails(db.Model):
    b = ForeignKeyField(BModel)


class APIKey(db.Model):
    key = CharField()
    secret = CharField()


class NotePanel(AdminPanel):
    template_name = 'admin/notes.html'
    
    def get_urls(self):
        return (
            ('/create/', self.create),
        )
    
    def create(self):
        if request.method == 'POST':
            if request.form.get('message'):
                Note.create(
                    user=auth.get_logged_in_user(),
                    message=request.form['message'],
                )
        next = request.form.get('next') or self.dashboard_url()
        return redirect(next)
    
    def get_context(self):
        return {
            'note_list': Note.select().order_by(('created_date', 'desc')).paginate(1, 3)
        }


auth = Auth(app, db, user_model=User)
admin = Admin(app, auth)


a_qf = QueryFilter(AModel.select())
bd_qf = QueryFilter(BDetails.select())
b_qf = QueryFilter(BModel.select(), related=[a_qf, bd_qf])
c_qf = QueryFilter(CModel.select(), related=[b_qf])
d_qf = QueryFilter(DModel.select(), related=[c_qf])

class AAdmin(ModelAdmin):
    columns = ('a_field',)

class BAdmin(ModelAdmin):
    columns = ('a', 'b_field',)
    related_filters = [a_qf]

class CAdmin(ModelAdmin):
    columns = ('b', 'c_field',)
    related_filters = [b_qf]

class DAdmin(ModelAdmin):
    columns = ('c', 'd_field',)
    related_filters = [c_qf]

class MessageAdmin(ModelAdmin):
    columns = ('user', 'content', 'pub_date',)

class NoteAdmin(ModelAdmin):
    columns = ('user', 'message', 'created_date',)


auth.register_admin(admin)
admin.register(AModel, AAdmin)
admin.register(BModel, BAdmin)
admin.register(CModel, CAdmin)
admin.register(DModel, DAdmin)
admin.register(BDetails)
admin.register(Message, MessageAdmin)
admin.register(Note, NoteAdmin)
admin.register_panel('Notes', NotePanel)


class UserResource(RestResource):
    exclude = ('password', 'email',)
    
    def get_query(self):
        return User.filter(active=True)


# rest api stuff
user_auth = UserAuthentication(auth)
admin_auth = AdminAuthentication(auth)
api_key_auth = APIKeyAuthentication(APIKey, ['GET', 'POST', 'PUT', 'DELETE'])

api = RestAPI(app, default_auth=user_auth)

api.register(Message, RestrictOwnerResource)
api.register(User, UserResource, auth=admin_auth)
api.register(Note)
api.register(TestModel, auth=api_key_auth)


# views
@app.route('/')
def homepage():
    return Response()

@app.route('/private/')
@auth.login_required
def private_timeline():
    return Response()


admin.setup()
api.setup()



import datetime
import operator

from flask import request
from flask_peewee.utils import models_to_path
from peewee import *
from wtforms import fields, form, widgets
from wtfpeewee.fields import ModelSelectField, ModelSelectMultipleField


class BooleanSelectField(fields.SelectFieldBase):
    widget = widgets.Select()

    def iter_choices(self):
        yield ('1', 'True', self.data)
        yield ('', 'False', not self.data)

    def process_data(self, value):
        try:
            self.data = bool(value)
        except (ValueError, TypeError):
            self.data = None

    def process_formdata(self, valuelist):
        if valuelist:
            try:
                self.data = bool(valuelist[0])
            except ValueError:
                raise ValueError(self.gettext(u'Invalid Choice: could not coerce'))


LOOKUP_TYPES = {
    'eq': 'equal to',
    'lt': 'less than',
    'gt': 'greater than',
    'lte': 'less than or equal to',
    'gte': 'greater than or equal to',
    'ne': 'not equal to',
    'istartswith': 'starts with',
    'icontains': 'contains',
    'in': 'is one of',
    
    # special lookups that require some preprocessing
    'today': 'is today',
    'yesterday': 'is yesterday',
    'this_week': 'this week',
    'lte_days_ago': 'less than ... days ago',
    'gte_days_ago': 'greater than ... days ago',
}

FIELD_TYPES = {
    'foreign_key': [ForeignKeyField],
    'text': [CharField, TextField],
    'numeric': [PrimaryKeyField, IntegerField, FloatField, DecimalField],
    'boolean': [BooleanField],
    'datetime': [DateTimeField],
}

INV_FIELD_TYPES = dict((v,k) for k in FIELD_TYPES for v in FIELD_TYPES[k])

FIELDS_TO_LOOKUPS = {
    'foreign_key': ['eq', 'in'],
    'text': ['eq', 'icontains', 'istartswith'],
    'numeric': ['eq', 'ne', 'lt', 'lte', 'gt', 'gte', 'in'],
    'boolean': ['eq'],
    'datetime': ['today', 'yesterday', 'this_week', 'lte_days_ago', 'gte_days_ago'],
}

CONVERTERS = {
    (ForeignKeyField, 'eq'): lambda f: ModelSelectField(model=f.to),
    (ForeignKeyField, 'in'): lambda f: ModelSelectMultipleField(model=f.to),
    (DateTimeField, 'today'): lambda f: fields.HiddenField(),
    (DateTimeField, 'yesterday'): lambda f: fields.HiddenField(),
    (DateTimeField, 'this_week'): lambda f: fields.HiddenField(),
    (BooleanField, 'eq'): lambda f: BooleanSelectField(),
}


def is_valid_lookup(model, lookup):
    lookups = lookup.split('__')
    lookup_type = 'eq'
    
    if len(lookups) > 1 and lookups[-1] in LOOKUP_TYPES:
        lookup_type = lookups.pop()
    
    curr = model
    for rel_field in lookups[:-1]:
        if rel_field in curr._meta.rel_fields:
            # is this a lookup using the descriptor name of a foreign key field,
            # if so change it to the actual field name, e.g. user -> user_id
            rel_field = curr._meta.rel_fields[rel_field]
        
        if rel_field in curr._meta.fields:
            field_obj = curr._meta.fields[rel_field]
            if isinstance(field_obj, ForeignKeyField):
                curr = field_obj.to
            else:
                return False
        else:
            if rel_field in curr._meta.reverse_relations:
                curr = curr._meta.reverse_relations[rel_field]
            else:
                return False
    
    lookup = lookups[-1]
    if lookup in curr._meta.rel_fields:
        lookup = curr._meta.rel_fields[lookup]

    if lookup not in curr._meta.fields:
        return False
    
    field_obj = curr._meta.fields[lookup]
    if lookup_type not in FIELDS_TO_LOOKUPS[INV_FIELD_TYPES[type(field_obj)]]:
        return False
    
    return (curr, lookup, lookup_type)

def lookups_for_field(field):
    """
    Get a list of valid lookups for a given field type.  Lookups are expressed
    as a 2-tuple of (<lookup suffix>, <human description>)
    
    >>> lookups_for_field(User.username)
    [('eq', 'equal to'), ('icontains', 'contains'), ('istartswith', 'starts with')]
    """
    field_class = type(field)
    return [
        (lookup, LOOKUP_TYPES[lookup]) \
            for lookup in FIELDS_TO_LOOKUPS[INV_FIELD_TYPES[field_class]]
    ]

class Lookup(object):
    def __init__(self, field):
        self.field = field
        
        self.field_class = type(self.field)
        self.field_name = self.field.name
        self.verbose_name = self.field.verbose_name
        self.model = field.model
    
    def __repr__(self):
        return '<Lookups for: %s.%s>' % (self.model.__name__, self.field_name)

    def __eq__(self, rhs):
        return self.field == rhs.field

    def get_field_type(self):
        return INV_FIELD_TYPES[self.field_class]
    
    def get_lookups(self):
        return lookups_for_field(self.field)
    
    def to_context(self):
        return dict(
            name=self.field_name,
            verbose_name=self.verbose_name,
            field_type=self.get_field_type(),
            lookups=self.get_lookups(),
            prefix=prefix,
        )
    
    def html_fields(self):
        default = lambda f: fields.TextField()
        return dict([
            (
                '%s__%s' % (self.field_name, lookup[0]),
                CONVERTERS.get((self.field_class, lookup[0]), default)(self.field)
            ) for lookup in self.get_lookups()
        ])

class ModelLookup(object):
    def __init__(self, model, exclude, path, raw_id_fields):
        self.model = model
        self.exclude = exclude
        self.path = path
        self.raw_id_fields = raw_id_fields
    
    def get_lookups(self):
        return [
            Lookup(f) for f in self.model._meta.get_fields() \
                if f.name not in self.exclude
        ]
    
    def get_prefix(self):
        if not self.path:
            return ''
        return models_to_path(self.path) + '__'
    
    def get_html_fields(self):
        fields = {}
        for lookup in self.get_lookups():
            fields.update(lookup.html_fields())
        return fields
    
    def html_form(self):
        frm = form.BaseForm(self.get_html_fields(), prefix=self.get_prefix())
        frm.process(None)
        return frm

def _rd(n):
    return datetime.date.today() + datetime.timedelta(days=n)


class FilterPreprocessor(object):
    def process_lookup(self, raw_lookup, values=None):
        """
        Returns a list of dictionaries
        """
        if '__' in raw_lookup:
            field_part, lookup = raw_lookup.rsplit('__', 1)
            if hasattr(self, 'process_%s' % lookup):
                return getattr(self, 'process_%s' % lookup)(field_part, values)
        
        return [{raw_lookup: v} for v in values]
    
    def process_in(self, field_part, values):
        return [{'%s__in' % field_part: values}]
    
    def process_today(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(0),
            '%s__lt' % field_part: _rd(1),
        }]
    
    def process_yesterday(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-1),
            '%s__lt' % field_part: _rd(0),
        }]
    
    def process_this_week(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-6),
            '%s__lt' % field_part: _rd(1),
        }]
    
    def process_lte_days_ago(self, field_part, values):
        return [{
            '%s__gte' % field_part: _rd(-1 * int(value)),
        } for value in values]
    
    def process_gte_days_ago(self, field_part, values):
        return [{
            '%s__lte' % field_part: _rd(-1 * int(value)),
        } for value in values]


class QueryFilter(object):
    def __init__(self, query, exclude_fields=None, ignore_filters=None, raw_id_fields=None, related=None):
        self.query = query
        self.model = self.query.model

        # fix any foreign key fields
        self.exclude_fields = []
        for field_name in exclude_fields or ():
            if field_name in self.model._meta.fields:
                self.exclude_fields.append(field_name)
            else:
                self.exclude_fields.append(self.model._meta.rel_fields[field_name])
        
        self.ignore_filters = ignore_filters or ()
        self.raw_id_fields = raw_id_fields or ()
        
        # a list of related QueryFilter() objects
        self.related = related or ()
    
    def get_model_lookups(self, path=None):
        model_lookups = [ModelLookup(self.model, self.exclude_fields, path, self.raw_id_fields)]
        path = path or []
        path.append(self.model)
        for related_filter in self.related:
            model_lookups.extend(
                related_filter.get_model_lookups(list(path))
            )
        return model_lookups
    
    def get_preprocessor(self):
        return FilterPreprocessor()
    
    def process_request(self):
        self.raw_lookups = []
        
        filters = []
        lookups_by_column = {}
        
        preprocessor = self.get_preprocessor()
        
        # preprocessing -- essentially a place to store "raw" filters (used to
        # reconstruct filter widgets on frontend), and a place to munge filters
        # that have special meaning, i.e. "today"
        for key in request.args:
            if key in self.ignore_filters:
                continue

            if not is_valid_lookup(self.model, key):
                continue
            
            values = request.args.getlist(key)
            if key.endswith('__in'):
                self.raw_lookups.append((key, values))
            else:
                for v in values:
                    self.raw_lookups.append((key, v))
            
            # preprocessor returns a list of filters to apply
            filters.extend(preprocessor.process_lookup(key, values))
        
        # at this point, need to figure out which parts of the query to "AND"
        # together and which to "OR" together ... this is naive and simply groups
        # lookups on the same column together in an OR clause.
        for filter_dict in filters:
            for field_part, value in filter_dict.items():
                if '__' in field_part:
                    column, lookup = field_part.rsplit('__', 1)
                else:
                    column = field_part
            
            lookups_by_column.setdefault(column, [])
            lookups_by_column[column].append(Q(**filter_dict))
        
        return [reduce(operator.or_, lookups) for lookups in lookups_by_column.values()]
    
    def get_filtered_query(self):
        nodes = self.process_request()
        if nodes:
            return self.query.filter(*nodes)
        
        return self.query
